#include "DirCodingRenderer.h"
#include "VectorAttributes.h"
#include "Grid.h"
#include <GLUT/glut.h>									//GLUT library
#include <algorithm>
#include <math.h>
#include <iostream>

using namespace std;



static float direction2angle(Point3d v)					//Converts a 2D vector into a direction (angle)
{														//The angle is normalized in the [0,1] interval 
	v.normalize();										//

	float cosa = v.x;
	float sina = v.y;
	
	float a;
	if (sina>=0)
		a = acos(cosa);
	else
		a = 2*M_PI - acos(cosa);
	
	return a/(2*M_PI);
}




static void hsv2rgb(float h, float s, float v, float& r, float& g, float& b)
{														//Convert from a HSV color to a RGB color
  float m, n, f; int i;

  h = 6*h;
  i = floor(h);
  f = h - i;
  if (!(i&1)) f = 1 - f; 
  m = v * (1 - s);
  n = v * (1 - s * f);

  switch(i) 
  {
	  case 6:
	  case 0: r=v; g=n; b=m; break; 
	  case 1: r=n; g=v; b=m; break; 
	  case 2: r=m; g=v; b=n; break; 
	  case 3: r=m; g=n; b=v; break; 
	  case 4: r=n; g=m; b=v; break; 
	  case 5: r=v; g=m; b=n; break; 
  } 
}



void DirCodingRenderer::draw(Grid& g)					//Draw 'g' by showing C1-interpolated vector-directions over cells
{
	glClearColor(1,1,1,1);								//0.   Clear the frame and depth buffers
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);			//1.   Render cells filled
	glDisable(GL_LIGHTING);
	glShadeModel(GL_SMOOTH);							//2.   Use bilinear interpolation when drawing cells as polygons		

	VectorAttributes& vectors = g.pointVectors();		//3.    Get vector attributes. We'll next compute and color-code their direction

	for(int i=0;i<g.numCells();++i)						//4. Draw all cells in the grid filled bilinearly-interpolated polygons
	{
		int   cell[10];									//Should have enough storage space for our largest cell
		
		int nvertices = g.getCell(i,cell);				//4.1. Get the 'nvertices' vertex-IDs of i-th cell
		if (nvertices!=3 && nvertices!=4)				//     We only handle here drawing of triangle and quad cells.
		   continue;									//     It is quite simple to extend this to other cell types.

		if (nvertices==3)								//     Triangle cells:
			glBegin(GL_TRIANGLES);						
		else //nvertices==4	
			glBegin(GL_QUADS);							//     Quad cells:
		
		for(int j=0;j<nvertices;++j)
		{
			float x[2];
			g.getPoint(cell[j],x);						//4.2. Get vertex coordinates of j-th vertex of i-th cell	
						
			const Point3d& v = g.pointVectors().getC0Vector(cell[j]);
			float direction = direction2angle(v);		//4.3. Get vector of current vertex. Convert it to a direction in [0,1]
			
			float R,G,B;								//4.4. Map direction to a color. We interpret direction as hue.
			hsv2rgb(direction,1,1,R,G,B);
			
			glColor3f(R,G,B);							//4.5. Pass vertex color to OpenGL
			glVertex3f(x[0],x[1],0);					//4.6. Pass vertex coordinate to OpenGL
		}	

		glEnd();                                        //4.7. Finish rendering current cell
	}
	
	glutSwapBuffers();
}


