#include "GlyphRenderer.h"
#include "ScalarAttributes.h"
#include "VectorAttributes.h"
#include "Grid.h"
#include <GLUT/glut.h>									//GLUT library
#include <algorithm>
#include <math.h>
#include <iostream>


using namespace std;


static float direction2angle(const Point3d& d)			//Converts a 2D vector into an orientation (angle).
{														//The angle is in the [0,360] degrees interval
	Point3d x = d; x.normalize();

	float cosa = x.x;
	float sina = x.y;
	
	float a;
	if (sina>=0)
		a = acos(cosa);
	else
		a = 2*M_PI - acos(cosa);
	
	return 180*a/M_PI;
}




void GlyphRenderer::drawMagnitude(Grid& g)				//Draw 'g' by showing vector-magnitude values by color mapping
{
	VectorAttributes& vectors = g.pointVectors();		//      Get vector range, to normalize vector lengths next in [0,1]
	float vector_min = vectors.getMin();
	float vector_max = vectors.getMax();
	
	glPointSize(3);										//1.    Draw all vertices in the grid as large points
	glBegin(GL_POINTS);									//2.    Draw vertices as an OpenGL point-set	
	for(int i=0;i<g.numPoints();++i)						
	{
		float v[2];
		g.getPoint(i,v);								//2.1.  Get coordinates of i-th vertex	
		
		Point3d vec = g.pointVectors().getC0Vector(i); 	//2.2.  Get	vector value at this vertex, and its normalized length too
		float length = (vec.norm()-vector_min)/(vector_max-vector_min);

		float r,g,b;									//
		mapToColor(length,r,g,b);						//2.3.  Map normalized length value to a color 
		glColor3f(r,g,b);								//2.4.  Pass color to OpenGL
		glVertex3f(v[0],v[1],0);						//2.5.  Pass this coordinate to OpenGL
	}
	glEnd();											//3.    Finish rendering the entire point-set
}


void GlyphRenderer::drawBalls(Grid& g)					//Draw 'g' by showing ball-glyphs, scaled to reflect the
{														//vector magnitude
	glEnable(GL_LIGHTING);
	glEnable(GL_LIGHT0);
	glLightModeli(GL_LIGHT_MODEL_TWO_SIDE, GL_TRUE);
	glEnable(GL_DEPTH_TEST);
	glEnable(GL_COLOR_MATERIAL);

	VectorAttributes& vectors = g.pointVectors();		//     Get vector range, to normalize vector lengths next in [0,1]
	float vector_min = vectors.getMin();
	float vector_max = vectors.getMax();
	
	glMatrixMode(GL_MODELVIEW);
	
	for(int i=0;i<g.numPoints();++i)						
	{
		float v[2];
		g.getPoint(i,v);								//2.1. Get coordinates of i-th vertex	
		
		Point3d vec = g.pointVectors().getC0Vector(i); 		
		float length = (vec.norm()-vector_min)/(vector_max-vector_min);

		float r,g,b;									//2.2.  Get vector-length value at i-th vertex and normalize it to [0,1]
		mapToColor(length,r,g,b);						//2.3.  Map normalized scalar value to a color 
		glColor3f(r,g,b);								//2.4.  Pass color to OpenGL

		float radius = length*scaling_factor/2;			//2.5.  Compute the radius at the current vertex

		glPushMatrix();
		glTranslatef(v[0],v[1],0);
		glutSolidSphere(radius,30,30);					//2.6.  Draw a ball glyph, centered at the current vertex, and scaled
		glPopMatrix();
	}

	glDisable(GL_LIGHTING);
}




void GlyphRenderer::drawCones(Grid& g)					//Draw 'g' by showing cone-glyphs, scaled and rotated to reflect the
{														//vector magnitude
	glEnable(GL_LIGHTING);
	glEnable(GL_LIGHT0);
	glLightModeli(GL_LIGHT_MODEL_TWO_SIDE, GL_TRUE);
	glEnable(GL_DEPTH_TEST);
	glEnable(GL_COLOR_MATERIAL);

	VectorAttributes& vectors = g.pointVectors();		//     Get vector range, to normalize vector lengths next in [0,1]
	float vector_min = vectors.getMin();
	float vector_max = vectors.getMax();
	
	glMatrixMode(GL_MODELVIEW);
	
	for(int i=0;i<g.numPoints();++i)						
	{
		float v[2];
		g.getPoint(i,v);								//2.1. Get coordinates of i-th vertex	
		
		Point3d vec = g.pointVectors().getC0Vector(i); 		
		
		float length = (vec.norm()-vector_min)/(vector_max-vector_min);

		float r,g,b;									//2.2.  Get vector-length value at i-th vertex and normalize it to [0,1]
		mapToColor(length,r,g,b);						//2.3.  Map normalized length to a color 
		glColor3f(r,g,b);								//2.4.  Pass color to OpenGL

		float scale = (scaling_mode==SCALE_NORMALIZE)? 1 : length; 
		float radius = scale*scaling_factor;			//2.5.  Compute the glyph's size, based on the vector length and scaling mode
		float angle = direction2angle(vec);				//2.6.  Convert the vector into an orientation (angle in [0,360] degrees)

		glPushMatrix();									//2.7.  Translate and rotate the canonical cone so as to be centered at the
		glTranslatef(v[0],v[1],0);						//      current vertex, and aligned with the current vector
		glRotatef(90-angle,0,0,-1);
		glRotatef(-90,1,0,0);
		glutSolidCone(radius/4,radius,30,30);			//2.8.  Draw the cone
		glPopMatrix();
	}

	glDisable(GL_LIGHTING);
}



void GlyphRenderer::drawArrows(Grid& g)					//Draw 'g' by showing line-glyphs, scaled to reflect the vector magnitude
{
	VectorAttributes& vectors = g.pointVectors();		//     Get vector range, to normalize vector lengths next in [0,1]
	float vector_min = vectors.getMin();
	float vector_max = vectors.getMax();
	
	
	glBegin(GL_LINES);									//1.   Draw line-glyphs as an OpenGL line-set	
	for(int i=0;i<g.numPoints();++i)						
	{
		Point3d v;
		g.getPoint(i,v.data);							//2.1. Get coordinates of i-th vertex	
		
		Point3d vec = g.pointVectors().getC0Vector(i); 		
		float length = (vec.norm()-vector_min)/(vector_max-vector_min);

		float r,g,b;									//2.2.  Get vector length value at i-th vertex and normalize it to [0,1]
		mapToColor(length,r,g,b);						//2.3.  Map normalized length value to a color 
		glColor3f(r,g,b);								//2.4.  Pass color to OpenGL
		
		vec.normalize();
		float scale = (scaling_mode==SCALE_NORMALIZE)? 1 : length; 
		vec *= scale*scaling_factor;
		Point3d tip = v + vec;							//2.5.  Compute 'tip' of the line-glyph

		glVertex3f(v[0],v[1],0);						//2.6.  Pass the two coordinates (current grid point and line tip)
		glVertex3f(tip[0],tip[1],0);					//      to OpenGL
	}
	glEnd();											//3.    Finish rendering the entire line-set
}



void GlyphRenderer::draw(Grid& g)						//Draw grid 'g' depending on drawing style attribute-value
{
	glClearColor(1,1,1,1);								//1. Clear the frame and depth buffers
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	switch (draw_style)
	{
	case DRAW_ARROWS: 
		drawArrows(g);
		break;
	case DRAW_MAGNITUDE: 
		drawMagnitude(g);
		break;
	case DRAW_BALLS:
		drawBalls(g);
		break;	
	case DRAW_CONES:
		drawCones(g);
		break;	
	}			
	
	glutSwapBuffers();
}




